unit DNNL;
// Intel® oneAPI Deep Neural Network Library (oneDNNL)
// Version 3.0 (2023.0)
(*******************************************************************************
 * Copyright 2016-2022 Intel Corporation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************)

{$IFDEF FPC}
  {$mode objfpc}
  {$align 4}
  {$packenum 4}
  {$packrecords C}
{$ENDIF}
{$INLINE ON}

interface

{$IFDEF WINDOWS}
  {$CALLING stdcall}
  const libdnnl = 'dnnl.dll';
{$ELSE}
  {$CALLING cdecl}
  const libdnnl = 'dnnl.so';
{$ENDIF}

(******************************************************************************
 *                  C API version definition (dnnl_version.h)                 *
 ******************************************************************************)
const
  /// Major version
  DNNL_VERSION_MAJOR = 3;
  /// Minor version
  DNNL_VERSION_MINOR = 0;

(******************************************************************************
 *                   C API types definitions (dnnl_types.h)                   *
 ******************************************************************************)
const
  /// Maximum number of dimensions a tensor can have. Only restricts the amount
  /// of space used for the tensor description. Individual computational
  /// primitives may support only tensors of certain dimensions.
  DNNL_MAX_NDIMS = 12;

  /// Maximum number of parts of RNN weights tensor that require separate
  /// computation.
  DNNL_RNN_MAX_N_PARTS = 4;

  /// Source argument #0.
  DNNL_ARG_SRC_0 = 1;
  /// A special mnemonic for source argument for primitives that have a
  /// single source. An alias for #DNNL_ARG_SRC_0.
  DNNL_ARG_SRC = DNNL_ARG_SRC_0;
  /// Source argument #1.
  DNNL_ARG_SRC_1 = 2;
  /// Destination argument #0.
  DNNL_ARG_DST_0 = 17;
  /// A special mnemonic for destination argument for primitives that have a
  /// single destination. An alias for #DNNL_ARG_DST_0.
  DNNL_ARG_DST = DNNL_ARG_DST_0;
  /// Weights argument #0.
  DNNL_ARG_WEIGHTS_0 = 33;
  /// A special mnemonic for primitives that have a single weights
  /// argument. Alias for #DNNL_ARG_WEIGHTS_0.
  DNNL_ARG_WEIGHTS = DNNL_ARG_WEIGHTS_0;
  /// Bias tensor argument.
  DNNL_ARG_BIAS = 41;
  /// Workspace tensor argument. Workspace is used to pass information
  /// from forward propagation to backward propagation computations.
  DNNL_ARG_WORKSPACE = 64;

  /// Gradient (diff) of the source argument #0.
  DNNL_ARG_DIFF_SRC_0 = 129;
  /// A special mnemonic for primitives that have a single diff source argument.
  /// An alias for #DNNL_ARG_DIFF_SRC_0.
  DNNL_ARG_DIFF_SRC = DNNL_ARG_DIFF_SRC_0;
  /// A special mnemonic for gradient (diff) of RNN input vector. An alias for
  /// #DNNL_ARG_DIFF_SRC_0.
  DNNL_ARG_DIFF_SRC_LAYER = DNNL_ARG_DIFF_SRC_0;

  /// Gradient (diff) of the destination argument #0.
  DNNL_ARG_DIFF_DST_0 = 145;
  /// A special mnemonic for primitives that have a single diff destination
  /// argument. An alias for #DNNL_ARG_DIFF_DST_0.
  DNNL_ARG_DIFF_DST = DNNL_ARG_DIFF_DST_0;
  /// A special mnemonic for gradient (diff) of RNN output vector. An alias for
  /// #DNNL_ARG_DIFF_DST_0.
  DNNL_ARG_DIFF_DST_LAYER = DNNL_ARG_DIFF_DST_0;
  /// Gradient (diff) of the weights argument #0.
  DNNL_ARG_DIFF_WEIGHTS_0 = 161;
  /// A special mnemonic for primitives that have a single diff weights
  /// argument. Alias for #DNNL_ARG_DIFF_WEIGHTS_0.
  DNNL_ARG_DIFF_WEIGHTS = DNNL_ARG_DIFF_WEIGHTS_0;
  /// Gradient (diff) of the bias tensor argument.
  DNNL_ARG_DIFF_BIAS = 169;
  /// Output scaling factors provided at execution time.
  DNNL_ARG_ATTR_OUTPUT_SCALES = 513;
  /// Starting index for source arguments for primitives that take a variable
  /// number of source arguments.
  DNNL_ARG_MULTIPLE_SRC = 1024;
  /// Scaling factors provided at execution time.
  DNNL_ARG_ATTR_SCALES = 4096;
  /// Zero points provided at execution time.
  DNNL_ARG_ATTR_ZERO_POINTS = 8192;

  /// Special pointer value that indicates that a memory object should not have
  /// an underlying buffer.
  DNNL_MEMORY_NONE = nil;

  /// Special pointer value that indicates that the library needs to allocate an
  /// underlying buffer for a memory object.
  DNNL_MEMORY_ALLOCATE = Pointer(-1);

type
  int = integer;
  unsigned = LongWord;
  float = single;
  size_t = PtrUInt;
  int32_t = int32;
  int64_t = int64;
  uint64_t = uint64;
  Pfloat = ^float;

(******************************************************************************
 *            C API common types definitions (dnnl_common_types.h)            *
 ******************************************************************************)
type
    /// Status values returned by the library functions.
    dnnl_status_t = (
    /// The operation was successful
    dnnl_success = 0,
    /// The operation failed due to an out-of-memory condition
    dnnl_out_of_memory = 1,
    /// The operation failed because of incorrect function arguments
    dnnl_invalid_arguments = 2,
    /// The operation failed because requested functionality is not implemented
    dnnl_unimplemented = 3,
    /// The last available implementation is reached
    dnnl_last_impl_reached = 4,
    /// Primitive or engine failed on execution
    dnnl_runtime_error = 5,
    /// Queried element is not required for given primitive
    dnnl_not_required = 6,
    /// The graph is not legitimate
    dnnl_invalid_graph = 7,
    /// The operation is not legitimate according to op schema
    dnnl_invalid_graph_op = 8,
    /// The shape cannot be inferred or compiled
    dnnl_invalid_shape = 9,
    /// The data type cannot be inferred or compiled
    dnnl_invalid_data_type = 10
  );

  /// Data type specification
  dnnl_data_type_t = (
    /// Undefined data type, used for empty memory descriptors.
    dnnl_data_type_undef = 0,
    /// 16-bit/half-precision floating point.
    dnnl_f16 = 1,
    /// non-standard 16-bit (bfloat16 w/ 7 bit mantissa) floating point.
    dnnl_bf16 = 2,
    /// 32-bit/single-precision floating point.
    dnnl_f32 = 3,
    /// 32-bit signed integer.
    dnnl_s32 = 4,
    /// 8-bit signed integer.
    dnnl_s8 = 5,
    /// 8-bit unsigned integer.
    dnnl_u8 = 6,
    /// 64-bit/double-precision floating point.
    dnnl_f64 = 7,

    /// Parameter to allow internal only data_types without undefined behavior.
    /// This parameter is chosen to be valid for so long as sizeof(int) >= 2.
    dnnl_data_type_max = $7fff
  );

  /// A type to describe tensor dimension.
  dnnl_dim_t = int64_t;
  Pdnnl_dim_t = ^dnnl_dim_t;
  /// A type to describe tensor dimensions.
  dnnl_dims_t = array[0..DNNL_MAX_NDIMS-1] of dnnl_dim_t;

  /// @brief Kinds of engines.
  dnnl_engine_kind_t = (
    /// An unspecified engine.
    dnnl_any_engine,
    /// CPU engine.
    dnnl_cpu,
    /// GPU engine.
    dnnl_gpu
  );
  Pdnnl_engine_kind_t = ^dnnl_engine_kind_t;

  /// @brief An opaque structure to describe an engine.
  dnnl_engine = record end;
  /// @brief An engine handle.
  dnnl_engine_t = ^dnnl_engine;
  Pdnnl_engine_t = ^dnnl_engine_t;

  /// @brief Stream flags.
  dnnl_stream_flags_t = (
    // In-order execution.
    dnnl_stream_in_order = $1,
    /// Out-of-order execution.
    dnnl_stream_out_of_order = $2,
    /// Default stream configuration.
    dnnl_stream_default_flags = dnnl_stream_in_order
  );

  /// @struct dnnl_stream
  /// An opaque structure to describe an execution stream.
  dnnl_stream = record end;
  /// An execution stream handle.
  dnnl_stream_t = ^dnnl_stream;
  Pdnnl_stream_t = ^dnnl_stream_t;
  /// A constant execution stream handle.
  const_dnnl_stream_t = ^dnnl_stream;

  /// Structure containing version information as per [Semantic
  /// Versioning](https://semver.org)
  dnnl_version_t = record
    major: int; ///< Major version
    minor: int; ///< Minor version
    patch: int; ///< Patch version
    hash: PAnsiChar; ///< Git hash of the sources (may be absent)
    cpu_runtime: unsigned; ///< CPU runtime
    gpu_runtime: unsigned; ///< GPU runtime
  end;
  Pdnnl_version_t = ^dnnl_version_t;


(******************************************************************************
 *                   C API types definitions (dnnl_types.h)                   *
 ******************************************************************************)
 type
  /// Memory format kind
  dnnl_format_kind_t = (
    /// Undefined memory format kind, used for empty memory descriptors.
    dnnl_format_kind_undef = 0,
    /// A special format kind that indicates that the actual format will be
    /// selected by a primitive automatically.
    dnnl_format_kind_any,
    /// A tensor in a generic format described by the stride and blocking
    /// values in each dimension.
    dnnl_blocked,
    /// A special format kind that indicates that tensor format is opaque.
    dnnl_format_kind_opaque,
    /// Parameter to allow internal only format kinds without undefined
    /// behavior. This parameter is chosen to be valid for so long as
    /// sizeof(int) >= 2.
    dnnl_format_kind_max = $7fff
  );

  /// Memory format tag specification.
  ///
  /// oneDNN formats describe physical data layout. The physical layout
  /// is described as a sequence of the dimensions as they are laid out in the
  /// memory (from the outer-most to the inner-most). Note that this order
  /// doesn't affect the logical order of the dimensions that is kept in the
  /// `dims` field of the dnnl_memory_desc_t structure. The logical order of the
  /// dimensions is specified by the primitive that uses the tensor.
  ///
  /// For example, CNN 5D tensor always has its logical dimensions in the order
  /// `(batch, channels, depth, height, width)`, while the physical layout might be
  /// `NCDHW` (corresponds to #dnnl_ncdhw format tag) or
  /// `NDHWC` (corresponds to #dnnl_ndhwc format tag).
  ///
  /// ~~~cpp
  /// int batch = 2, channels = 16, depth = 13, height = 13, width = 13;
  ///
  /// int ndims = 5; // 5D tensor
  /// dnnl_dims_t dims = {batch, channels, depth, height, width};
  /// dnnl_memory_desc_t data_in_ncdhw;
  /// dnnl_memory_desc_create_with_tag(
  ///      &data_in_ncdhw, 5, dims, dnnl_f32, dnnl_ncdhw);
  ///
  /// // note that in both cases dims passed are the same
  /// dnnl_memory_desc_t data_in_ndhwc;
  /// dnnl_memory_desc_create_with_tag(
  ///      &data_in_ndhwc, 5, dims, dnnl_f32, dnnl_ndhwc);
  ///
  /// dnnl_memory_desc_destroy(data_in_ncdhw);
  /// dnnl_memory_desc_destroy(data_in_ndhwc);
  /// ~~~
  ///
  /// Memory format tags can be further divided into two categories:
  ///  - Domain-agnostic names, i.e. names the do not depend on the tensor usage
  ///    in the specific primitive. These names use letters from `a` to `l` to
  ///    denote logical dimension from 1 to 12, and form the order in which the
  ///    dimensions are laid in memory. For instance, #dnnl_ab is used to denote
  ///    2D tensor where the second logical dimension (aka `b`) is the innermost,
  ///    i.e. has stride = 1, and the first logical dimension (`a`) laid out in
  ///    memory with stride equal to the size of second dimension. On the other
  ///    hand, #dnnl_ba is just transposed version of the same tensor: the
  ///    first dimension (`a`) becomes the innermost one.
  ///  - Domain-specific names, i.e. names that make sense only in the context of
  ///    a certain domain, such as CNN. This names are just aliases to the
  ///    corresponding domain-agnostic tags and used mostly for the convenience.
  ///    For example, #dnnl_nc is used to denote 2D CNN activations tensor
  ///    memory format, where channels are the innermost dimension and batch is an
  ///    outermost one. Moreover, #dnnl_nc is just an alias to #dnnl_ab,
  ///    since for oneDNN CNN primitives the logical dimensions of
  ///    activations tensors come in order: batch, channels, spatial.
  ///    In other words, batch corresponds to the first logical dimension (`a`),
  ///    channels correspond to the second one (`b`).
  ///
  /// The following domain-specific notation applies to memory format tags:
  ///  - @c 'n' denotes the mini-batch dimension
  ///  - @c 'c' denotes a channels dimension
  ///  - When there are multiple channel dimensions (for example, in convolution
  ///    weights tensor), @c 'i' and @c 'o' denote dimensions of input and output
  ///    channels
  ///  - @c 'd', @c 'h', and @c 'w' denote spatial depth, height, and width
  ///    respectively
  ///
  /// Upper-case letters indicate that the data is laid out in blocks for a
  /// particular dimension. In such cases, the format name contains both upper-
  /// and lower-case letters for that dimension with a lower-case letter preceded
  /// by the block size. For example: #dnnl_nChw8c describes a format where the
  /// outermost dimension is mini-batch, followed by the channel block number,
  /// followed by the spatial height and width, and finally followed by 8-element
  /// channel blocks.
  ///
  /// @sa @ref dev_guide_understanding_memory_formats
  dnnl_format_tag_t = (
    /// Undefined memory format tag
    dnnl_format_tag_undef = 0,
    /// Undefined memory format tag.
    /// The primitive selects a format automatically.
    dnnl_format_tag_any,

    // Semantic agnostic section
    // The physical order of dimensions is defined by the permutation of the
    // characters, assuming that ab..z defines the natural order.

    // Plain formats

    dnnl_a, ///< plain 1D tensor
    dnnl_ab, ///< plain 2D tensor
    dnnl_abc, ///< plain 3D tensor
    dnnl_abcd, ///< plain 4D tensor
    dnnl_abcde, ///< plain 5D tensor
    dnnl_abcdef, ///< plain 6D tensor
    dnnl_abcdefg, ///< plain 7D tensor
    dnnl_abcdefgh, ///< plain 8D tensor
    dnnl_abcdefghi, ///< plain 9D tensor
    dnnl_abcdefghij, ///< plain 10D tensor
    dnnl_abcdefghijk, ///< plain 11D tensor
    dnnl_abcdefghijkl, ///< plain 12D tensor

    // Permuted plain formats

    dnnl_ba, ///< permuted 2D tensor
    dnnl_acb, ///< permuted 3D tensor
    dnnl_bac, ///< permuted 3D tensor
    dnnl_bca, ///< permuted 3D tensor
    dnnl_cab, ///< permuted 3D tensor
    dnnl_cba, ///< permuted 3D tensor
    dnnl_abdc, ///< permuted 4D tensor
    dnnl_acbd, ///< permuted 4D tensor
    dnnl_acdb, ///< permuted 4D tensor
    dnnl_adbc, ///< permuted 4D tensor
    dnnl_adcb, ///< permuted 4D tensor
    dnnl_bacd, ///< permuted 4D tensor
    dnnl_bcda, ///< permuted 4D tensor
    dnnl_cdab, ///< permuted 4D tensor
    dnnl_cdba, ///< permuted 4D tensor
    dnnl_dcab, ///< permuted 4D tensor
    dnnl_abced, ///< permuted 5D tensor
    dnnl_abdec, ///< permuted 5D tensor
    dnnl_acbde, ///< permuted 5D tensor
    dnnl_acdeb, ///< permuted 5D tensor
    dnnl_adecb, ///< permuted 5D tensor
    dnnl_bacde, ///< permuted 5D tensor
    dnnl_bcdea, ///< permuted 5D tensor
    dnnl_cdeab, ///< permuted 5D tensor
    dnnl_cdeba, ///< permuted 5D tensor
    dnnl_decab, ///< permuted 5D tensor
    dnnl_abcdfe, ///< permuted 6D tensor
    dnnl_abdefc, ///< permuted 6D tensor
    dnnl_abdfce, ///< permuted 6D tensor
    dnnl_acbdef, ///< permuted 6D tensor
    dnnl_adefcb, ///< permuted 6D tensor
    dnnl_defcab, ///< permuted 6D tensor
    dnnl_abcdegf, ///< permuted 7D tensor
    dnnl_abcdefhg, ///< permuted 8D tensor
    dnnl_abcdefgih, ///< permuted 9D tensor
    dnnl_abcdefghji, ///< permuted 10D tensor
    dnnl_abcdefghikj, ///< permuted 11D tensor
    dnnl_abcdefghijlk, ///< permuted 12D tensor

    // Opaque blocked formats

    dnnl_Abc16a,
    dnnl_ABc16a16b,
    dnnl_ABc32a32b,
    dnnl_ABc4a4b,
    /// 3D tensor blocked by 2nd dimension with block size 16
    dnnl_aBc16b,
    dnnl_ABc16b16a,
    dnnl_Abc4a,
    /// 3D tensor blocked by 2nd dimension with block size 32
    dnnl_aBc32b,
    /// 3D tensor blocked by 2nd dimension with block size 4
    dnnl_aBc4b,
    dnnl_ABc4b16a4b,
    dnnl_ABc2b8a4b,
    dnnl_ABc16b16a4b,
    dnnl_ABc16b16a2b,
    dnnl_ABc4b4a,
    dnnl_ABc8a16b2a,
    dnnl_ABc8a8b,
    dnnl_ABc8a4b,
    /// 3D tensor blocked by 2nd dimension with block size 8
    dnnl_aBc8b,
    dnnl_ABc8b16a2b,
    dnnl_BAc8a16b2a,
    dnnl_ABc8b8a,
    dnnl_Abcd16a,
    dnnl_Abcd8a,
    dnnl_ABcd16a16b,
    dnnl_Abcd32a,
    dnnl_ABcd32a32b,
    /// 4D tensor blocked by 2nd dimension with block size 16
    dnnl_aBcd16b,
    dnnl_ABcd16b16a,
    dnnl_aBCd16b16c,
    dnnl_aBCd16c16b,
    dnnl_Abcd4a,
    /// 4D tensor blocked by 2nd dimension with block size 32
    dnnl_aBcd32b,
    /// 4D tensor blocked by 2nd dimension with block size 4
    dnnl_aBcd4b,
    dnnl_ABcd4b16a4b,
    dnnl_ABcd16b16a4b,
    dnnl_ABcd16b16a2b,
    dnnl_ABcd4b4a,
    dnnl_ABcd4a4b,
    dnnl_aBCd2c4b2c,
    dnnl_aBCd4b8c2b,
    dnnl_aBCd4c16b4c,
    dnnl_aBCd2c8b4c,
    dnnl_aBCd16c16b4c,
    dnnl_aBCd16c16b2c,
    dnnl_aBCd4c4b,
    dnnl_aBCd4b4c,
    dnnl_ABcd8a16b2a,
    dnnl_ABcd2b8a4b,
    dnnl_ABcd8a8b,
    dnnl_ABcd8a4b,
    /// 4D tensor blocked by 2nd dimension with block size 8
    dnnl_aBcd8b,
    dnnl_aBCd4c8b2c,
    dnnl_ABcd8b16a2b,
    dnnl_aBCd8b16c2b,
    dnnl_BAcd8a16b2a,
    /// 4D tensor blocked by 1st and 2nd dimension with block size 8
    dnnl_ABcd8b8a,
    dnnl_aBCd8b8c,
    dnnl_aBCd8b4c,
    dnnl_aBCd8c16b2c,
    dnnl_ABcde8a16b2a,
    dnnl_aCBd8b16c2b,
    dnnl_aBCd8c8b,
    dnnl_Abcde16a,
    dnnl_Abcde32a,
    dnnl_ABcde16a16b,
    dnnl_BAcde8a16b2a,
    /// 4D tensor blocked by 3rd dimension with block size 4
    dnnl_aBCd2b4c2b,
    /// 5D tensor blocked by 1st dimension with block size 16
    dnnl_ABcde4b16a4b,
    /// 5D tensor blocked by 1st dimension with block size 8
    dnnl_ABcde2b8a4b,
    /// 5D tensor blocked by 2nd dimension with block size 16
    dnnl_aBcde16b,
    dnnl_ABcde16b16a,
    dnnl_aBCde16b16c,
    dnnl_aBCde16c16b,
    dnnl_aBCde2c8b4c,
    dnnl_Abcde4a,
    /// 5D tensor blocked by 2nd dimension with block size 32
    dnnl_aBcde32b,
    /// 5D tensor blocked by 2nd dimension with block size 4
    dnnl_aBcde4b,
    dnnl_ABcde4b4a,
    dnnl_ABcde4a4b,
    dnnl_aBCde4b4c,
    dnnl_aBCde2c4b2c,
    dnnl_aBCde4b8c2b,
    dnnl_aBCde4c16b4c,
    dnnl_aBCde16c16b4c,
    dnnl_aBCde16c16b2c,
    dnnl_aBCde4c4b,
    dnnl_Abcde8a,
    dnnl_ABcde8a8b,
    dnnl_ABcde8a4b,
    dnnl_BAcde16b16a,
    /// 5D tensor blocked by 2nd dimension with block size 8
    dnnl_aBcde8b,
    dnnl_ABcde8b16a2b,
    dnnl_aBCde8b16c2b,
    dnnl_aBCde4c8b2c,
    dnnl_aCBde8b16c2b,
    dnnl_ABcde8b8a,
    dnnl_ABcde32a32b,
    dnnl_aBCde8b8c,
    dnnl_aBCde8b4c,
    dnnl_ABc4a8b8a4b,
    dnnl_ABcd4a8b8a4b,
    dnnl_ABcde4a8b8a4b,
    dnnl_BAc4b8a8b4a,
    dnnl_BAcd4b8a8b4a,
    dnnl_BAcde4b8a8b4a,
    dnnl_ABcd2a8b8a2b,
    dnnl_aBCd4b8c8b4c,
    dnnl_aBCde4b8c8b4c,
    dnnl_aBCde2b8c8b2c,
    dnnl_aBCde8c16b2c,
    dnnl_aBCde8c8b,
    /// 5D tensor blocked by 3rd dimension with block size 4
    dnnl_aBCde2b4c2b,
    /// 6D tensor blocked by 2nd dimension with block size 16
    dnnl_aBcdef16b,
    dnnl_aBCdef16b16c,
    dnnl_aBCdef16c16b,
    dnnl_aBCdef4c16b4c,
    /// 6D tensor blocked by 2nd dimension with block size 8
    dnnl_aBCdef2c8b4c,
    dnnl_aBCdef4c8b2c,
    /// 6D tensor blocked by 3rd dimension with block size 4
    dnnl_aBCdef2b4c2b,
    /// 6D tensor blocked by 2nd dimension with block size 4
    dnnl_aBcdef4b,
    dnnl_aBCdef4c4b,
    dnnl_aBCdef4b4c,
    dnnl_aBCdef2c4b2c,
    dnnl_aBCdef4b8c2b,
    dnnl_aBCdef8b8c,
    dnnl_aBCdef8b4c,
    dnnl_aBCdef8c16b2c,
    dnnl_aBCdef4b8c8b4c,
    dnnl_aBCdef8b16c2b,
    dnnl_aCBdef8b16c2b,
    dnnl_aBCdef8c8b,
    dnnl_aBdc16b,
    dnnl_aBdC16b2c,
    dnnl_aBdC16b4c,
    dnnl_aBdc4b,
    dnnl_aBdc8b,
    dnnl_aBdec16b,
    dnnl_aBdeC16b2c,
    dnnl_aBdeC16b4c,
    dnnl_aBdec32b,
    dnnl_aBdec4b,
    dnnl_aBdec8b,
    dnnl_aBdefc16b,
    dnnl_aBdefC16b2c,
    dnnl_aCBdef16c16b,
    dnnl_aBdefc4b,
    dnnl_aBdefc8b,
    dnnl_Abcdef16a,
    dnnl_Abcdef32a,
    dnnl_aBedc16b,
    dnnl_Acb16a,
    dnnl_AcB16a2b,
    dnnl_AcB16a4b,
    dnnl_Acb4a,
    dnnl_Acb8a,
    dnnl_aCBd16b16c,
    dnnl_aCBd16c16b,
    dnnl_aCBde16b16c,
    dnnl_aCBde16c16b,
    dnnl_Acdb16a,
    dnnl_AcdB16a2b,
    dnnl_AcdB16a4b,
    dnnl_Acdb32a,
    dnnl_Acdb4a,
    dnnl_Acdb8a,
    dnnl_Acdeb16a,
    dnnl_AcdeB16a2b,
    dnnl_Acdeb4a,
    dnnl_Acdeb8a,
    dnnl_Adcb16a,
    dnnl_BAc16a16b,
    dnnl_BAc16b16a,
    dnnl_BAcd16a16b,
    dnnl_BAcd16b16a,
    dnnl_aCBd4c8b8c4b,
    dnnl_aCBde4c8b8c4b,
    dnnl_aCBdef4c8b8c4b,
    dnnl_BAcde16a16b,
    dnnl_aCBdef16b16c,
    dnnl_ABc16b32a,
    dnnl_ABc16b64a,
    dnnl_ABc4b32a4b,
    dnnl_ABc4b64a4b,
    dnnl_ABc8b32a2b,
    dnnl_ABc8b64a2b,
    dnnl_AB16b16a,
    dnnl_AB16b32a,
    dnnl_AB16b64a,
    dnnl_AB8b16a2b,
    dnnl_AB8b32a2b,
    dnnl_AB8b64a2b,
    dnnl_AB4b16a4b,
    dnnl_AB4b32a4b,
    dnnl_AB4b64a4b,
    dnnl_AB16b16a4b,
    dnnl_ABcd16b32a,
    dnnl_ABcd16b64a,
    dnnl_ABcd4b32a4b,
    dnnl_ABcd4b64a4b,
    dnnl_ABcd8b32a2b,
    dnnl_ABcd8b64a2b,
    dnnl_ABcde4b32a4b,
    dnnl_ABcde4b64a4b,
    dnnl_ABcde16b16a4b,
    dnnl_ABcde16b16a2b,
    dnnl_ABcde16b32a,
    dnnl_ABcde16b64a,
    dnnl_ABcde8b32a2b,
    dnnl_ABcde8b64a2b,
    dnnl_aBCdef16c16b4c,
    dnnl_aBCdef16c16b2c,
    dnnl_AB32a32b8a4b,
    dnnl_AB8a4b,
    dnnl_AB32a32b8a2b,
    dnnl_AB8a2b,
    dnnl_abDc32d,
    dnnl_abDC32d4c,
    dnnl_abdEc32e,
    dnnl_abdEC32e2c,
    dnnl_abdEC32e4c,
    dnnl_aBdefC16b4c,
    dnnl_AcdeB16a4b,
    dnnl_ABcd16a16b2a,
    dnnl_ABc16a16b2a,
    dnnl_aBCd16b16c2b,
    dnnl_aBCde16b16c2b,
    dnnl_Acb32a,
    dnnl_AcB32a2b,
    dnnl_AcB32a4b,
    dnnl_Acb48a,
    dnnl_AcB48a2b,
    dnnl_AcB48a4b,
    dnnl_Acb64a,
    dnnl_AcB64a2b,
    dnnl_AcB64a4b,
    dnnl_cBa2b,
    dnnl_cBa4b,
    dnnl_aBdc32b,
    dnnl_aBdC32b2c,
    dnnl_aBdC32b4c,
    dnnl_aBdc48b,
    dnnl_aBdC48b2c,
    dnnl_aBdC48b4c,
    dnnl_aBdc64b,
    dnnl_aBdC64b2c,
    dnnl_aBdC64b4c,
    dnnl_adCb2c,
    dnnl_adCb4c,
    dnnl_AcdB32a2b,
    dnnl_AcdB32a4b,
    dnnl_Acdb48a,
    dnnl_AcdB48a2b,
    dnnl_AcdB48a4b,
    dnnl_Acdb64a,
    dnnl_AcdB64a2b,
    dnnl_AcdB64a4b,
    dnnl_cdBa2b,
    dnnl_cdBa4b,
    dnnl_aBdeC32b2c,
    dnnl_aBdeC32b4c,
    dnnl_aBdec48b,
    dnnl_aBdeC48b2c,
    dnnl_aBdeC48b4c,
    dnnl_aBdec64b,
    dnnl_aBdeC64b2c,
    dnnl_aBdeC64b4c,
    dnnl_adeCb2c,
    dnnl_adeCb4c,
    dnnl_Acdeb32a,
    dnnl_AcdeB32a2b,
    dnnl_AcdeB32a4b,
    dnnl_Acdeb48a,
    dnnl_AcdeB48a2b,
    dnnl_AcdeB48a4b,
    dnnl_Acdeb64a,
    dnnl_AcdeB64a2b,
    dnnl_AcdeB64a4b,
    dnnl_cdeBa2b,
    dnnl_cdeBa4b,
    dnnl_aBdefc32b,
    dnnl_aBdefC32b2c,
    dnnl_aBdefC32b4c,
    dnnl_aBdefc48b,
    dnnl_aBdefC48b2c,
    dnnl_aBdefC48b4c,
    dnnl_aBdefc64b,
    dnnl_aBdefC64b2c,
    dnnl_aBdefC64b4c,
    dnnl_adefCb2c,
    dnnl_adefCb4c,
    dnnl_AB16b32a4b,
    dnnl_AB16b48a4b,
    dnnl_AB16b64a4b,
    dnnl_AB16b16a2b,
    dnnl_AB16b32a2b,
    dnnl_AB16b48a2b,
    dnnl_AB16b64a2b,
    dnnl_ABc16b32a4b,
    dnnl_ABc16b48a4b,
    dnnl_ABc16b64a4b,
    dnnl_ABc16b32a2b,
    dnnl_ABc16b48a2b,
    dnnl_ABc16b64a2b,
    dnnl_ABcd16b32a4b,
    dnnl_ABcd16b48a4b,
    dnnl_ABcd16b64a4b,
    dnnl_ABcd16b32a2b,
    dnnl_ABcd16b48a2b,
    dnnl_ABcd16b64a2b,
    dnnl_ABcde16b32a4b,
    dnnl_ABcde16b48a4b,
    dnnl_ABcde16b64a4b,
    dnnl_ABcde16b32a2b,
    dnnl_ABcde16b48a2b,
    dnnl_ABcde16b64a2b,
    dnnl_ABc32a16b,
    dnnl_ABcd32a16b,
    dnnl_ABcde32a16b,
    dnnl_AB48a16b,
    dnnl_AB48a32b,
    dnnl_ABc40a16b,
    dnnl_ABc40a32b,
    dnnl_aBC48b16c,
    dnnl_aBC48b32c,
    dnnl_ABcd40a16b,
    dnnl_ABcd40a32b,
    dnnl_abCd32c,
    dnnl_abdCe32c,
    dnnl_abdCE32c2e,
    dnnl_BA16a16b2a,
    dnnl_BA16a32b2a,
    dnnl_BA16a48b2a,
    dnnl_BA16a64b2a,
    dnnl_BA16a16b4a,
    dnnl_BA16a32b4a,
    dnnl_BA16a48b4a,
    dnnl_BA16a64b4a,
    dnnl_ABcd8a2b,
    dnnl_aBdeC16c16b2c,
    dnnl_aBdeC16c16b4c,
    dnnl_aBdefC16c16b2c,
    dnnl_AcB16b16a2b,
    dnnl_AcB16b16a4b,
    dnnl_AcdB16b16a2b,
    dnnl_AcdB16b16a4b,
    dnnl_AcdeB16b16a2b,
    dnnl_aBdefC16c16b4c,
    dnnl_AcdeB16b16a4b,
    dnnl_AcB16b32a2b,
    dnnl_AcB16b32a4b,
    dnnl_AcB16b48a2b,
    dnnl_AcB16b48a4b,
    dnnl_AcB16b64a2b,
    dnnl_AcB16b64a4b,
    dnnl_aBdC16c16b2c,
    dnnl_aBdC16c16b4c,
    dnnl_aBdC16c32b2c,
    dnnl_aBdC16c32b4c,
    dnnl_aBdC16c48b2c,
    dnnl_aBdC16c48b4c,
    dnnl_aBdC16c64b2c,
    dnnl_aBdC16c64b4c,
    dnnl_AcdB16b32a2b,
    dnnl_AcdB16b32a4b,
    dnnl_AcdB16b48a2b,
    dnnl_AcdB16b48a4b,
    dnnl_AcdB16b64a2b,
    dnnl_AcdB16b64a4b,
    dnnl_aBdeC16c32b2c,
    dnnl_aBdeC16c32b4c,
    dnnl_aBdeC16c48b2c,
    dnnl_aBdeC16c48b4c,
    dnnl_aBdeC16c64b2c,
    dnnl_aBdeC16c64b4c,
    dnnl_AcdeB16b32a2b,
    dnnl_AcdeB16b32a4b,
    dnnl_AcdeB16b48a2b,
    dnnl_AcdeB16b48a4b,
    dnnl_AcdeB16b64a2b,
    dnnl_AcdeB16b64a4b,
    dnnl_aBdefC16c32b2c,
    dnnl_aBdefC16c32b4c,
    dnnl_aBdefC16c48b2c,
    dnnl_aBdefC16c48b4c,
    dnnl_aBdefC16c64b2c,
    dnnl_aBdefC16c64b4c,
    dnnl_decbA16a,
    dnnl_ABc4a2b,
    dnnl_ABc8a2b,
    dnnl_aBCd8b2c,
    dnnl_ABcde4a2b,
    dnnl_ABcde8a2b,
    dnnl_ABcde40a16b,
    dnnl_ABcde40a32b,
    dnnl_aBCde8b2c,
    dnnl_ABcde4a8b8a2b,
    dnnl_ABcd4a8b8a2b,
    dnnl_ABc4a8b8a2b,
    dnnl_aBCdef4b8c8b2c,
    dnnl_aBCde4b8c8b2c,
    dnnl_aBCd4b8c8b2c,
    dnnl_BAcde4b8a8b2a,
    dnnl_BAcd4b8a8b2a,
    dnnl_BAc4b8a8b2a,
    dnnl_aCBdef4c8b8c2b,
    dnnl_aCBde4c8b8c2b,
    dnnl_aCBd4c8b8c2b,
    dnnl_aBCdef8b2c,
    dnnl_AB32a16b,
    dnnl_AB32a32b,
    dnnl_BA4b8a8b2a,
    dnnl_BA4b8a8b4a,
    dnnl_aBC32b16c,
    dnnl_aBC32b32c,
    dnnl_aCB4c8b8c2b,
    dnnl_aCB4c8b8c4b,
    dnnl_ABcd4a2b,
    dnnl_ABc2b8a16b4a,
    dnnl_ABcd2b8a16b4a,
    dnnl_ABcde2b8a16b4a,
    dnnl_ABc2a8b16a4b,
    dnnl_ABc2a8b16a2b,
    dnnl_ABc2b32a8b,
    dnnl_ABcd2a8b16a4b,
    dnnl_ABcd2a8b16a2b,
    dnnl_aCBd2c8b16c2b,
    dnnl_ABcd2b32a8b,
    dnnl_aBCd2c8b16c2b,
    dnnl_ABcde2a8b16a4b,
    dnnl_ABcde2a8b16a2b,
    dnnl_aCBde2c8b16c2b,
    dnnl_ABcde2b32a8b,
    dnnl_aBC2b8c16b2c,
    dnnl_aBCd2b8c16b2c,
    dnnl_aBCde2b8c16b2c,
    dnnl_aBCdef2b8c16b2c,
    dnnl_BAcde2b8a16b4a,
    dnnl_BAcd2b8a16b4a,
    dnnl_BAc2b8a16b4a,
    dnnl_BAcde2b8a16b2a,
    dnnl_BAcd2b8a16b2a,
    dnnl_BAc2b8a16b2a,
    dnnl_aBCde2c8b16c2b,
    dnnl_aBCdef2c8b16c2b,
    dnnl_aCBdef2c8b16c2b,
    dnnl_aBCd2b8c16b4c,
    dnnl_aBCde2b8c16b4c,
    dnnl_BA4b8a16b2a,
    dnnl_BA4b8a16b4a,
    dnnl_aCB4c8b16c2b,
    dnnl_aCB4c8b16c4b,
    dnnl_BA16a16b,
    dnnl_BA16a32b,
    dnnl_BA16a48b,
    dnnl_BA16a64b,
    dnnl_aCB16c2b,
    dnnl_aCB16c4b,
    dnnl_BA16b2a,
    dnnl_BA16b4a,
    dnnl_aBC16b16c,
    dnnl_aBC16b32c,
    dnnl_AB16a16b,
    dnnl_AB16a32b,
    dnnl_ABcde16a16b2a,
    dnnl_aBCdef16b16c2b,
    dnnl_Acedb16a,
    dnnl_aBdfec16b,
    dnnl_abdEC64e2c,
    dnnl_abdEC64e4c,
    dnnl_aCB16b16c,
    dnnl_aCB16b32c,
    dnnl_aCB16b48c,
    dnnl_aCB16b64c,
    dnnl_aCB16b16c2b,
    dnnl_aCB16b32c2b,
    dnnl_aCB16b48c2b,
    dnnl_aCB16b64c2b,
    dnnl_aCB16b16c4b,
    dnnl_aCB16b32c4b,
    dnnl_aCB16b48c4b,
    dnnl_aCB16b64c4b,
    dnnl_abCd4c,
    dnnl_abCde4c,
    dnnl_abCdef4c,
    dnnl_abCde32c,
    dnnl_abCdef32c,
    dnnl_ABcd16a32b,
    dnnl_decbA8a,
    dnnl_aCdefB16b32c2b,
    dnnl_aCdefB16b32c4b,
    dnnl_aCdefB16b48c2b,
    dnnl_aCdefB16b48c4b,
    dnnl_aCdefB16b64c2b,
    dnnl_aCdefB16b64c4b,
    dnnl_BcdeA16a32b2a,
    dnnl_BcdeA16a32b4a,
    dnnl_BcdeA16a48b2a,
    dnnl_BcdeA16a48b4a,
    dnnl_BcdeA16a64b2a,
    dnnl_BcdeA16a64b4a,
    dnnl_aCdefb32c,
    dnnl_aCdefB32c2b,
    dnnl_aCdefB32c4b,
    dnnl_aCdefb48c,
    dnnl_aCdefB48c2b,
    dnnl_aCdefB48c4b,
    dnnl_aCdefb64c,
    dnnl_aCdefB64c2b,
    dnnl_aCdefB64c4b,
    dnnl_Bcdea32b,
    dnnl_BcdeA32b2a,
    dnnl_BcdeA32b4a,
    dnnl_Bcdea48b,
    dnnl_BcdeA48b2a,
    dnnl_BcdeA48b4a,
    dnnl_Bcdea64b,
    dnnl_BcdeA64b2a,
    dnnl_BcdeA64b4a,
    dnnl_Bca32b,
    dnnl_BcA32b2a,
    dnnl_BcA32b4a,
    dnnl_Bca48b,
    dnnl_BcA48b2a,
    dnnl_BcA48b4a,
    dnnl_Bca64b,
    dnnl_BcA64b2a,
    dnnl_BcA64b4a,
    dnnl_aCdb32c,
    dnnl_aCdB32c2b,
    dnnl_aCdB32c4b,
    dnnl_aCdb48c,
    dnnl_aCdB48c2b,
    dnnl_aCdB48c4b,
    dnnl_aCdb64c,
    dnnl_aCdB64c2b,
    dnnl_aCdB64c4b,
    dnnl_BcA16a16b2a,
    dnnl_BcA16a16b4a,
    dnnl_BcdA16a16b2a,
    dnnl_BcdA16a16b4a,
    dnnl_BcdeA16a16b2a,
    dnnl_BcdeA16a16b4a,
    dnnl_aCdB16b16c2b,
    dnnl_aCdB16b16c4b,
    dnnl_aCdeB16b16c2b,
    dnnl_aCdeB16b16c4b,
    dnnl_aCdefB16b16c2b,
    dnnl_aCdefB16b16c4b,
    dnnl_BcA16a32b2a,
    dnnl_BcA16a32b4a,
    dnnl_BcA16a48b2a,
    dnnl_BcA16a48b4a,
    dnnl_BcA16a64b2a,
    dnnl_BcA16a64b4a,
    dnnl_aCdB16b32c2b,
    dnnl_aCdB16b32c4b,
    dnnl_aCdB16b48c2b,
    dnnl_aCdB16b48c4b,
    dnnl_aCdB16b64c2b,
    dnnl_aCdB16b64c4b,
    dnnl_BcdA16a32b2a,
    dnnl_BcdA16a32b4a,
    dnnl_BcdA16a48b2a,
    dnnl_BcdA16a48b4a,
    dnnl_BcdA16a64b2a,
    dnnl_BcdA16a64b4a,
    dnnl_aCdeB16b32c2b,
    dnnl_aCdeB16b32c4b,
    dnnl_aCdeB16b48c2b,
    dnnl_aCdeB16b48c4b,
    dnnl_aCdeB16b64c2b,
    dnnl_aCdeB16b64c4b,
    dnnl_Bca16b,
    dnnl_BcA16b2a,
    dnnl_BcA16b4a,
    dnnl_Bcda16b,
    dnnl_BcdA16b2a,
    dnnl_BcdA16b4a,
    dnnl_Bcdea16b,
    dnnl_BcdeA16b2a,
    dnnl_BcdeA16b4a,
    dnnl_aCdb16c,
    dnnl_aCdB16c2b,
    dnnl_aCdB16c4b,
    dnnl_aCdeb16c,
    dnnl_aCdeB16c2b,
    dnnl_aCdeB16c4b,
    dnnl_aCdefb16c,
    dnnl_aCdefB16c2b,
    dnnl_aCdefB16c4b,
    dnnl_Bcda32b,
    dnnl_BcdA32b2a,
    dnnl_BcdA32b4a,
    dnnl_Bcda48b,
    dnnl_BcdA48b2a,
    dnnl_BcdA48b4a,
    dnnl_Bcda64b,
    dnnl_BcdA64b2a,
    dnnl_BcdA64b4a,
    dnnl_aCdeb32c,
    dnnl_aCdeB32c2b,
    dnnl_aCdeB32c4b,
    dnnl_aCdeb48c,
    dnnl_aCdeB48c2b,
    dnnl_aCdeB48c4b,
    dnnl_aCdeb64c,
    dnnl_aCdeB64c2b,
    dnnl_aCdeB64c4b,

    /// Just a sentinel, not real memory format tag. Must be changed after new
    /// format tag is added.
    dnnl_format_tag_last,

    // Aliases

    /// 1D tensor, an alias to #dnnl_a
    dnnl_x = dnnl_a,
    /// 2D CNN activations tensor, an alias to #dnnl_ab
    dnnl_nc = dnnl_ab,
    /// 2D CNN activations tensor, an alias to #dnnl_ba
    dnnl_cn = dnnl_ba,
    /// 2D RNN statistics tensor, an alias to #dnnl_ab
    dnnl_tn = dnnl_ab,
    /// 2D RNN statistics tensor, an alias to #dnnl_ba
    dnnl_nt = dnnl_ba,
    /// 3D CNN activations tensor, an alias to #dnnl_abc
    dnnl_ncw = dnnl_abc,
    /// 3D CNN activations tensor, an alias to #dnnl_acb
    dnnl_nwc = dnnl_acb,
    /// 4D CNN activations tensor, an alias to #dnnl_abcd
    dnnl_nchw = dnnl_abcd,
    /// 4D CNN activations tensor, an alias to #dnnl_acdb
    dnnl_nhwc = dnnl_acdb,
    /// 4D CNN activations tensor, an alias to #dnnl_bcda
    dnnl_chwn = dnnl_bcda,
    /// 5D CNN activations tensor, an alias to #dnnl_abcde
    dnnl_ncdhw = dnnl_abcde,
    /// 5D CNN activations tensor, an alias to #dnnl_acdeb
    dnnl_ndhwc = dnnl_acdeb,

    /// 2D CNN weights tensor, an alias to #dnnl_ab
    dnnl_oi = dnnl_ab,
    /// 2D CNN weights tensor, an alias to #dnnl_ba
    dnnl_io = dnnl_ba,
    /// 3D CNN weights tensor, an alias to #dnnl_abc
    dnnl_oiw = dnnl_abc,
    /// 3D CNN weights tensor, an alias to #dnnl_acb
    dnnl_owi = dnnl_acb,
    /// 3D CNN weights tensor, an alias to #dnnl_cba
    dnnl_wio = dnnl_cba,
    /// 3D CNN weights tensor, an alias to #dnnl_cab
    dnnl_woi = dnnl_cab,
    /// 3D CNN weights tensor, an alias to #dnnl_bca
    dnnl_iwo = dnnl_bca,
    /// 4D CNN weights tensor, an alias to #dnnl_abcd
    dnnl_oihw = dnnl_abcd,
    /// 4D CNN weights tensor, an alias to #dnnl_cdba
    dnnl_hwio = dnnl_cdba,
    /// 4D CNN weights tensor, an alias to #dnnl_cdab
    dnnl_hwoi = dnnl_cdab,
    /// 4D CNN weights tensor, an alias to #dnnl_acdb
    dnnl_ohwi = dnnl_acdb,
    /// 4D CNN weights tensor, an alias to #dnnl_bcda
    dnnl_ihwo = dnnl_bcda,
    /// 4D CNN weights tensor, an alias to #dnnl_bacd
    dnnl_iohw = dnnl_bacd,
    /// 5D CNN weights tensor, an alias to #dnnl_abcde
    dnnl_oidhw = dnnl_abcde,
    /// 5D CNN weights tensor, an alias to #dnnl_bacde
    dnnl_iodhw = dnnl_bacde,
    /// 5D CNN weights tensor, an alias to #dnnl_cdeba
    dnnl_dhwio = dnnl_cdeba,
    /// 5D CNN weights tensor, an alias to #dnnl_cdeab
    dnnl_dhwoi = dnnl_cdeab,
    /// 5D CNN weights tensor, an alias to #dnnl_acdeb
    dnnl_odhwi = dnnl_acdeb,
    /// 5D CNN weights tensor, an alias to #dnnl_bcdea
    dnnl_idhwo = dnnl_bcdea,

    /// 4D CNN weights tensor (incl. groups), an alias to #dnnl_abcd
    dnnl_goiw = dnnl_abcd,
    /// 4D CNN weights tensor (incl. groups), an alias to #dnnl_abdc
    dnnl_gowi = dnnl_abdc,
    /// 4D CNN weights tensor (incl. groups), an alias to #dnnl_dcab
    dnnl_wigo = dnnl_dcab,
    /// 5D CNN weights tensor (incl. groups), an alias to #dnnl_abcde
    dnnl_goihw = dnnl_abcde,
    /// 5D CNN weights tensor (incl. groups), an alias to #dnnl_abdec
    dnnl_gohwi = dnnl_abdec,
    /// 5D CNN weights tensor (incl. groups), an alias to #dnnl_decab
    dnnl_hwigo = dnnl_decab,
    /// 5D CNN weights tensor (incl. groups), an alias to #dnnl_acbde
    dnnl_giohw = dnnl_acbde,
    /// 6D CNN weights tensor (incl. groups), an alias to #dnnl_abcdef
    dnnl_goidhw = dnnl_abcdef,
    /// 6D CNN weights tensor (incl. groups), an alias to #dnnl_abdefc
    dnnl_godhwi = dnnl_abdefc,
    /// 6D CNN weights tensor (incl. groups), an alias to #dnnl_acbdef
    dnnl_giodhw = dnnl_acbdef,
    /// 6D CNN weights tensor (incl. groups), an alias to #dnnl_defcab
    dnnl_dhwigo = dnnl_defcab,

    /// 3D RNN data tensor in the format (seq_length, batch, input channels),
    /// an alias to #dnnl_abc.
    dnnl_tnc = dnnl_abc,
    /// 3D RNN data tensor in the format (batch, seq_length, input channels),
    /// an alias to #dnnl_bac.
    dnnl_ntc = dnnl_bac,
    /// 4D RNN states tensor in the format (num_layers, num_directions,
    /// batch, state channels), an alias to #dnnl_abcd.
    dnnl_ldnc = dnnl_abcd,
    /// 5D RNN weights tensor in the format (num_layers, num_directions,
    /// input_channels, num_gates, output_channels), an alias to #dnnl_abcde.
    ///
    ///  - For LSTM cells, the gates order is input, forget, candidate
    ///    and output gate.
    ///  - For GRU cells, the gates order is update, reset and output gate.
    dnnl_ldigo = dnnl_abcde,
    /// 5D RNN weights tensor in the format (num_layers, num_directions,
    /// num_gates, output_channels, input_channels), an alias to #dnnl_abdec.
    ///
    ///  - For LSTM cells, the gates order is input, forget, candidate
    ///    and output gate.
    ///  - For GRU cells, the gates order is update, reset and output gate.
    dnnl_ldgoi = dnnl_abdec,
    /// 4D LSTM projection tensor in the format (num_layers, num_directions,
    /// num_channels_in_hidden_state, num_channels_in_recurrent_projection),
    /// an alias to #dnnl_abcd.
    dnnl_ldio = dnnl_abcd,
    /// 4D LSTM projection tensor in the format (num_layers, num_directions,
    /// num_channels_in_recurrent_projection, num_channels_in_hidden_state),
    /// an alias to #dnnl_abdc.
    dnnl_ldoi = dnnl_abdc,
    /// 4D RNN bias tensor in the format (num_layers, num_directions,
    /// num_gates, output_channels), an alias to #dnnl_abcd.
    ///
    ///  - For LSTM cells, the gates order is input, forget, candidate
    ///    and output gate.
    ///  - For GRU cells, the gates order is update, reset and output gate.
    dnnl_ldgo = dnnl_abcd,
    /// 5D LSTM projection tensor
    dnnl_ldOi32o = dnnl_abDc32d,
    dnnl_ldOI32o4i = dnnl_abDC32d4c,
    dnnl_ldIo32i = dnnl_abCd32c,
    /// 6D RNN weights tensor
    dnnl_ldgOi32o = dnnl_abdEc32e,
    dnnl_ldgOI32o2i = dnnl_abdEC32e2c,
    dnnl_ldgOI32o4i = dnnl_abdEC32e4c,
    dnnl_ldgOI64o2i = dnnl_abdEC64e2c,
    dnnl_ldgOI64o4i = dnnl_abdEC64e4c,
    dnnl_ldgIo32i = dnnl_abdCe32c,
    dnnl_ldgIO32i2o = dnnl_abdCE32c2e
  );

  /// Kinds of propagation.
  dnnl_prop_kind_t = (
    // TODO: suggest renames
    /// Undefined propagation type.
    dnnl_prop_kind_undef = 0,
    /// Forward data propagation (training mode). In this mode primitives
    /// perform computations necessary for subsequent backward propagation.
    dnnl_forward_training = 64,
    /// Forward data propagation (inference mode). In this mode primitives
    /// perform only computations that are necessary for inference and omit
    /// computations that are necessary only for backward propagation.
    dnnl_forward_inference = 96,
    /// Forward data propagation (alias for @c dnnl_forward_training).
    dnnl_forward = dnnl_forward_training,
    /// Backward propagation (with respect to all parameters).
    dnnl_backward = 128,
    /// Backward data propagation.
    dnnl_backward_data = 160,
    /// Backward weights propagation.
    dnnl_backward_weights = 192,
    /// Backward bias propagation.
    dnnl_backward_bias = 193
  );

  /// Kinds of primitives. Used to implement a way to extend the library with new
  /// primitives without changing the ABI.
  dnnl_primitive_kind_t = (
    /// Undefined primitive
    dnnl_undefined_primitive,
    /// A reorder primitive.
    dnnl_reorder,
    /// A shuffle primitive.
    dnnl_shuffle,
    /// A (out-of-place) concat primitive.
    dnnl_concat,
    /// A sum primitive.
    dnnl_sum,
    /// A convolution primitive.
    dnnl_convolution,
    /// A deconvolution primitive.
    dnnl_deconvolution,
    /// An element-wise primitive.
    dnnl_eltwise,
    /// An LRN primitive.
    dnnl_lrn,
    /// A batch normalization primitive.
    dnnl_batch_normalization,
    /// An inner product primitive.
    dnnl_inner_product,
    /// A rnn primitive.
    dnnl_rnn,
    /// A matrix multiplication primitive (internal).
    dnnl_gemm,
    /// A binary primitive.
    dnnl_binary,
    /// A matrix multiplication primitive.
    dnnl_matmul,
    /// A resampling primitive.
    dnnl_resampling,
    /// A pooling primitive.
    dnnl_pooling,
    /// A reduction primitive.
    dnnl_reduction,
    /// A PReLU primitive.
    dnnl_prelu,
    /// A softmax primitive.
    dnnl_softmax,
    /// A layer normalization primitive.
    dnnl_layer_normalization,

    /// Parameter to allow internal only primitives without undefined behavior.
    /// This parameter is chosen to be valid for so long as sizeof(int) >= 2.
    dnnl_primitive_kind_max = $7fff
  );

  /// Kinds of algorithms.
  dnnl_alg_kind_t = (
    dnnl_alg_kind_undef,
    /// Direct convolution
    dnnl_convolution_direct = $1,
    /// Winograd convolution
    dnnl_convolution_winograd = $2,
    /// Convolution algorithm(either direct or Winograd) is chosen just in time
    dnnl_convolution_auto = $3,
    /// Direct deconvolution
    dnnl_deconvolution_direct = $a,
    /// Winograd deconvolution
    dnnl_deconvolution_winograd = $b,
    /// Eltwise: ReLU
    dnnl_eltwise_relu = $20,
    /// Eltwise: hyperbolic tangent non-linearity (tanh)
    dnnl_eltwise_tanh,
    /// Eltwise: exponential linear unit (elu)
    dnnl_eltwise_elu,
    /// Eltwise: square
    dnnl_eltwise_square,
    /// Eltwise: abs
    dnnl_eltwise_abs,
    /// Eltwise: square root
    dnnl_eltwise_sqrt,
    /// Eltwise: linear
    dnnl_eltwise_linear,
    /// Eltwise: soft_relu
    dnnl_eltwise_soft_relu,
    /// Eltwise: hardsigmoid
    dnnl_eltwise_hardsigmoid,
    /// Eltwise: logistic
    dnnl_eltwise_logistic,
    /// Eltwise: exponent
    dnnl_eltwise_exp,
    /// Eltwise: gelu
    ///
    /// @note Tanh approximation formula is used to approximate
    /// the cumulative distribution function of a Gaussian here
    dnnl_eltwise_gelu_tanh,
    /// Eltwise: swish
    dnnl_eltwise_swish,
    /// Eltwise: natural logarithm
    dnnl_eltwise_log,
    /// Eltwise: clip
    dnnl_eltwise_clip,
    /// Eltwise: clip version 2
    dnnl_eltwise_clip_v2,
    /// Eltwise: pow
    dnnl_eltwise_pow,
    /// Eltwise: erf-based gelu
    dnnl_eltwise_gelu_erf,
    /// Eltwise: round
    dnnl_eltwise_round,
    /// Eltwise: mish
    dnnl_eltwise_mish,
    /// Eltwise: hardswish
    dnnl_eltwise_hardswish,
    /// Eltwise: ReLU (dst for backward)
    dnnl_eltwise_relu_use_dst_for_bwd = $100,
    /// Eltwise: hyperbolic tangent non-linearity (tanh) (dst for backward)
    dnnl_eltwise_tanh_use_dst_for_bwd,
    /// Eltwise: exponential linear unit (elu) (dst for backward)
    dnnl_eltwise_elu_use_dst_for_bwd,
    /// Eltwise: square root (dst for backward)
    dnnl_eltwise_sqrt_use_dst_for_bwd,
    /// Eltwise: logistic (dst for backward)
    dnnl_eltwise_logistic_use_dst_for_bwd,
    /// Eltwise: exp (dst for backward)
    dnnl_eltwise_exp_use_dst_for_bwd,
    /// Eltwise: clip version 2 (dst for backward)
    dnnl_eltwise_clip_v2_use_dst_for_bwd,
    /// Max pooling
    dnnl_pooling_max = $1ff,
    /// Average pooling include padding
    dnnl_pooling_avg_include_padding = $2ff,
    /// Average pooling exclude padding
    dnnl_pooling_avg_exclude_padding = $3ff,
    /// Local response normalization (LRN) across multiple channels
    dnnl_lrn_across_channels = $aff,
    /// LRN within a single channel
    dnnl_lrn_within_channel = $bff,
    /// RNN cell
    dnnl_vanilla_rnn = $1fff,
    /// LSTM cell
    dnnl_vanilla_lstm = $2fff,
    /// GRU cell
    dnnl_vanilla_gru = $3fff,
    /// GRU cell with linear before reset
    ///
    /// Modification of original GRU cell. Differs from #dnnl_vanilla_gru
    /// in how the new memory gate is calculated:
    /// \f[ c_t = tanh(W_c*x_t + b_{c_x} + r_t*(U_c*h_{t-1}+b_{c_h})) \f]
    /// Primitive expects 4 biases on input:
    /// \f$[b_{u}, b_{r}, b_{c_x}, b_{c_h}]\f$
    dnnl_lbr_gru = $4fff,
    /// AUGRU cell
    dnnl_vanilla_augru = $5fff,
    /// AUGRU cell with linear before reset
    dnnl_lbr_augru = $6fff,
    /// Binary add
    dnnl_binary_add = $1fff0,
    /// Binary mul
    dnnl_binary_mul = $1fff1,
    /// Binary max
    dnnl_binary_max = $1fff2,
    /// Binary min
    dnnl_binary_min = $1fff3,
    /// Binary div
    dnnl_binary_div = $1fff4,
    /// Binary sub
    dnnl_binary_sub = $1fff5,
    /// Binary greater or equal
    dnnl_binary_ge = $1fff6,
    /// Binary greater than
    dnnl_binary_gt = $1fff7,
    /// Binary less or equal
    dnnl_binary_le = $1fff8,
    /// Binary less than
    dnnl_binary_lt = $1fff9,
    /// Binary equal
    dnnl_binary_eq = $1fffa,
    /// Binary not equal
    dnnl_binary_ne = $1fffb,
    /// Nearest Neighbor Resampling Method
    dnnl_resampling_nearest = $2fff0,
    /// Linear Resampling Method
    dnnl_resampling_linear = $2fff1,
    /// Reduction using max
    dnnl_reduction_max,
    /// Reduction using min
    dnnl_reduction_min,
    /// Reduction using sum
    dnnl_reduction_sum,
    /// Reduction using mul
    dnnl_reduction_mul,
    /// Reduction using mean
    dnnl_reduction_mean,
    /// Reduction using lp norm
    dnnl_reduction_norm_lp_max,
    /// Reduction using lp norm
    dnnl_reduction_norm_lp_sum,
    /// Reduction using lp norm without final pth-root
    dnnl_reduction_norm_lp_power_p_max,
    /// Reduction using lp norm without final pth-root
    dnnl_reduction_norm_lp_power_p_sum,
    /// Softmax
    dnnl_softmax_accurate = $30000,
    /// Logsoftmax
    dnnl_softmax_log
  );

  /// @struct dnnl_memory_desc
  /// An opaque structure to describe a memory descriptor.
  dnnl_memory_desc = record end;
  /// A memory descriptor handle.
  dnnl_memory_desc_t = ^dnnl_memory_desc;
  Pdnnl_memory_desc_t = ^dnnl_memory_desc_t;
  /// A memory descriptor handle.
  const_dnnl_memory_desc_t = ^dnnl_memory_desc;
  Pconst_dnnl_memory_desc_t = ^const_dnnl_memory_desc_t;

  /// @struct dnnl_memory
  /// An opaque structure to describe a memory.
  dnnl_memory = record end;

  /// A memory handle.
  dnnl_memory_t = ^dnnl_memory;
  Pdnnl_memory_t = ^dnnl_memory_t;

  /// A constant memory handle.
  const_dnnl_memory_t = ^dnnl_memory;

  /// @brief A primitive descriptor handle.
  dnnl_primitive_desc = record end;
  dnnl_primitive_desc_t = ^dnnl_primitive_desc;
  Pdnnl_primitive_desc_t = ^dnnl_primitive_desc_t;

  /// @brief A constant primitive descriptor handle.
  const_dnnl_primitive_desc_t = ^dnnl_primitive_desc;
  Pconst_dnnl_primitive_desc_t = ^const_dnnl_primitive_desc_t;

  /// @struct dnnl_primitive_attr
  /// @brief An opaque structure for primitive descriptor attributes.
  ///
  /// Attributes may contain:
  ///  - output scales (to scale the result prior to storing it to the memory)
  dnnl_primitive_attr = record end;
  dnnl_primitive_attr_t = ^dnnl_primitive_attr;
  Pdnnl_primitive_attr_t = ^dnnl_primitive_attr_t;

  /// @brief A constant primitive descriptor attributes handle.
  const_dnnl_primitive_attr_t = ^dnnl_primitive_attr;
  Pconst_dnnl_primitive_attr_t = ^const_dnnl_primitive_attr_t;

  /// @struct dnnl_post_ops
  /// @brief An opaque structure for a chain of post operations.
  ///
  /// dnnl_post_ops can be used to perform some (trivial) operations like
  /// accumulation or eltwise after certain primitives like convolution.
  ///
  /// Post operations might be combined together, making a chain of post
  /// operations. For instance one can configure convolution followed by
  /// accumulation followed by eltwise. This might be especially beneficial
  /// for residual learning blocks.
  ///
  /// @warning
  ///      Of course not all combinations are supported, so the user should handle
  ///      errors accordingly.
  ///
  /// Supported post operations:
  ///  - accumulation (base primitive: convolution)
  ///  - eltwise (base primitive: convolution)
  dnnl_post_ops = record end;

  /// @brief A post operation chain handle.
  dnnl_post_ops_t = ^dnnl_post_ops;
  Pdnnl_post_ops_t = ^dnnl_post_ops_t;

  /// @brief A constant post operation chain handle.
  const_dnnl_post_ops_t = ^dnnl_post_ops;
  Pconst_dnnl_post_ops_t = ^const_dnnl_post_ops_t;

  /// @struct dnnl_primitive
  /// An opaque structure to describe a primitive.
  dnnl_primitive = record end;
  /// A primitive handle.
  dnnl_primitive_t = ^dnnl_primitive;
  Pdnnl_primitive_t = ^dnnl_primitive_t;
  /// A constant primitive handle.
  const_dnnl_primitive_t = ^dnnl_primitive;

  /// A structure that contains an index and a memory object, and is used to pass
  /// arguments to dnnl_primitive_execute().
  dnnl_exec_arg_t = record
    arg: int;              ///< An argument index, e.g. DNNL_ARG_SRC
    memory: dnnl_memory_t; ///< Input/output memory
  end;
  Pdnnl_exec_arg_t = ^dnnl_exec_arg_t;

  /// Primitive descriptor query specification
  ///
  /// @note
  ///     Rule of thumb: all opaque types and structures are returned by
  ///     reference. All numbers are returned by value.
  ///
  /// @warning
  ///     All returned references point to constant objects and are valid only
  ///     during the lifetime of the queried primitive descriptor. Returned objects
  ///     must not be destroyed by the user. If you need to keep the object longer
  ///     than the lifetime of the queried primitive descriptor, use
  ///     dnnl_primitive_desc_clone() to make a copy.
  dnnl_query_t = (
    dnnl_query_undef = 0, ///< no query

    dnnl_query_engine, ///< execution engine
    dnnl_query_primitive_kind, ///< primitive kind

    dnnl_query_num_of_inputs_s32, ///< number of inputs expected
    dnnl_query_num_of_outputs_s32, ///< number of outputs expected

    dnnl_query_time_estimate_f64, ///< runtime estimation (seconds)
    dnnl_query_memory_consumption_s64, ///< memory consumption -- extra
    ///  (scratch) memory, additional to
    ///  all inputs and outputs memory
    ///  (bytes)

    dnnl_query_scratchpad_engine, ///< scratchpad engine -- engine to be used
    ///  for creating scratchpad memory

    dnnl_query_impl_info_str, ///< implementation name

    dnnl_query_reorder_src_engine, ///< source engine
    dnnl_query_reorder_dst_engine, ///< destination engine

    dnnl_query_prop_kind, ///< propagation kind

    dnnl_query_cache_blob_id_size_s64, ///< size of cache blob ID in bytes
    dnnl_query_cache_blob_id, ///< cache blob  ID (pointer to array)

    dnnl_query_strides, ///< strides
    dnnl_query_dilations, ///< dilations
    dnnl_query_padding_l, ///< left padding
    dnnl_query_padding_r, ///< right padding
    dnnl_query_epsilon_f32, ///< epsilon
    dnnl_query_flags, ///< flags
    dnnl_query_alg_kind, ///< algorithm kind
    dnnl_query_alpha_f32, ///< alpha
    dnnl_query_beta_f32, ///< beta
    dnnl_query_axis_s32, ///< axis
    dnnl_query_local_size_s64, ///< LRN parameter local size
    dnnl_query_k_f32, ///< LRN parameter K
    dnnl_query_p_f32, ///< Reduction parameter P
    dnnl_query_factors, ///< Resampling parameter factors
    dnnl_query_cell_kind, ///< RNN parameter cell kind
    dnnl_query_direction, ///< RNN parameter direction
    dnnl_query_activation_kind, ///< RNN parameter activation kind
    dnnl_query_kernel, ///< Pooling parameter kernel
    dnnl_query_group_size_s64, ///< Shuffle parameter group size

    // memory descriptor section
    dnnl_query_some_md = 128, ///< stub
    dnnl_query_src_md, ///< source memory desc
    dnnl_query_diff_src_md, ///< source gradient memory desc
    dnnl_query_weights_md, ///< weights memory descriptor desc
    dnnl_query_diff_weights_md, ///< weights grad. memory desc
    dnnl_query_dst_md, ///< destination memory desc
    dnnl_query_diff_dst_md, ///< destination grad. memory desc
    dnnl_query_workspace_md, ///< workspace memory desc
    dnnl_query_scratchpad_md, ///< scratchpad memory desc
    dnnl_query_exec_arg_md = 255, ///< memory desc of an execute argument

    dnnl_query_ndims_s32, ///< number of dimensions
    dnnl_query_dims, ///< vector of dimensions
    dnnl_query_data_type, ///< data type
    dnnl_query_submemory_offset_s64, ///< submemory offset
    dnnl_query_padded_dims, ///< vector of padded dimensions
    dnnl_query_padded_offsets, ///< vector of padded offsets
    dnnl_query_format_kind, ///< format kind
    dnnl_query_inner_nblks_s32, ///< number of innermost blocks
    dnnl_query_inner_blks, ///< vector of sizes of the innermost blocks
    dnnl_query_inner_idxs, ///< vector of logical indices of the blocks

    // Max value to prevent UB for internal use only dnnl_query_t
    dnnl_query_max = $7fff
  );


(******************************************************************************
 *                        C common API (dnnl_common.h)                        *
 ******************************************************************************)

/// Returns the number of engines of a particular kind.
///
function dnnl_engine_get_count(kind: dnnl_engine_kind_t): size_t; external libdnnl;

/// Creates an engine.
///
/// @param engine Output engine.
/// @param kind Engine kind.
/// @param index Engine index that should be between 0 and the count of
///     engines of the requested kind.
/// @returns #dnnl_success on success and a status describing the error
///     otherwise.
function dnnl_engine_create(engine: Pdnnl_engine_t; kind: dnnl_engine_kind_t; index: size_t): dnnl_status_t; external libdnnl;

/// Returns the kind of an engine.
///
function dnnl_engine_get_kind(engine: dnnl_engine_t; kind: Pdnnl_engine_kind_t): dnnl_status_t; external libdnnl;

/// Destroys an engine.
///
/// @param engine Engine to destroy.
/// @returns #dnnl_success on success and a status describing the error
///     otherwise.
function dnnl_engine_destroy(engine: dnnl_engine_t): dnnl_status_t; external libdnnl;

/// Creates an execution stream.
///
function dnnl_stream_create(stream: Pdnnl_stream_t; engine: dnnl_engine_t; flags: unsigned): dnnl_status_t; external libdnnl;

/// Waits for all primitives in the execution stream to finish computations.
///
function dnnl_stream_wait(stream: dnnl_stream_t): dnnl_status_t; external libdnnl;

/// Destroys an execution stream.
///
function dnnl_stream_destroy(stream: dnnl_stream_t): dnnl_status_t; external libdnnl;

/// Configures verbose output to stdout.
///
/// @note
///     Enabling verbose output affects performance.
///     This setting overrides the ONEDNN_VERBOSE environment variable.
///
function dnnl_set_verbose(level: int): dnnl_status_t; external libdnnl;

/// Returns library version information.
function dnnl_version(): Pdnnl_version_t; external libdnnl;


(******************************************************************************
 *                               C API (dnnl.h)                               *
 ******************************************************************************)

/// Clones a primitive descriptor. The resulting primitive descriptor must be
/// destroyed separately.
///
function dnnl_primitive_desc_clone(
        primitive_desc: Pdnnl_primitive_desc_t;
        existing_primitive_desc: const_dnnl_primitive_desc_t): dnnl_status_t; external libdnnl;

/// Returns a constant reference to the attributes of a primitive descriptor.
///
/// @warning
///     It is an error to destroy the resulting @p attr.
///
/// @warning
///     The lifetime of an @p attr is the same as that of a @p
///     primitive_desc, so it is an error to use the @p attr once the @p
///     primitive_desc has been destroyed.
///
function dnnl_primitive_desc_get_attr(
        primitive_desc: const_dnnl_primitive_desc_t;
        attr: Pconst_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Destroys a primitive descriptor.
///
function dnnl_primitive_desc_destroy(primitive_desc: dnnl_primitive_desc_t): dnnl_status_t; external libdnnl;

/// Queries a primitive descriptor for various pieces of information.
///
/// The most common use case is to query a primitive descriptor, created with
/// source, weights, and destination memory descriptors with format tags set
/// to #dnnl_format_tag_any, for the corresponding memory descriptors (in this
/// case the @p what is set to #dnnl_query_src_md, #dnnl_query_weights_md, and
/// #dnnl_query_dst_md respectively) so that it is possible to create memory
/// objects and reorder primitives if necessary.
///
function dnnl_primitive_desc_query(
        primitive_desc: const_dnnl_primitive_desc_t; what: dnnl_query_t;
        index: int; result: Pointer): dnnl_status_t; external libdnnl;

/// Queries primitive descriptor for a memory descriptor.
///
/// @note
///     This function is a convenience version of
///     #dnnl_primitive_desc_query().
///
function dnnl_primitive_desc_query_md(primitive_desc: const_dnnl_primitive_desc_t;
        what: dnnl_query_t; index: int): const_dnnl_memory_desc_t; external libdnnl;

/// Creates a primitive.
///
function dnnl_primitive_create(primitive: Pdnnl_primitive_t; primitive_desc: const_dnnl_primitive_desc_t): dnnl_status_t; external libdnnl;

/// Executes a primitive.
///
function dnnl_primitive_execute(primitive: const_dnnl_primitive_t;
        stream: dnnl_stream_t; nargs: int; const args: Pdnnl_exec_arg_t): dnnl_status_t; external libdnnl;

/// Retrieves a constant reference to the primitive descriptor of a given
/// primitive.
///
/// @warning
///     It is an error to destroy the returned object. It is owned by the
///     primitive. The @c const qualifier of the returned object prevents
///     such attempts.
///
function dnnl_primitive_get_primitive_desc(
        primitive: const_dnnl_primitive_t;
        primitive_desc: Pconst_dnnl_primitive_desc_t): dnnl_status_t; external libdnnl;

/// Destroys a primitive.
///
function dnnl_primitive_destroy(primitive: dnnl_primitive_t): dnnl_status_t; external libdnnl;


/// Creates an empty (default) primitive attributes with all the parameters
/// set to their default values.
///
/// Empty attributes are implied whenever the respective argument is NULL.
///
function dnnl_primitive_attr_create(attr: Pdnnl_primitive_attr_t): dnnl_status_t; external libdnnl;


/// Destroys primitive attributes.
///
/// @param attr Primitive attributes to destroy.
/// @returns #dnnl_success on success and a status describing the error
///     otherwise.
function dnnl_primitive_attr_destroy(attr: dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Sets primitive attributes scaling factors for primitive operations for a
/// given memory argument. The scaling factors must be passed at execution time
/// as an argument with index #DNNL_ARG_ATTR_SCALES | arg.
///
function dnnl_primitive_attr_set_scales_mask(
        attr: dnnl_primitive_attr_t; arg: int; mask: int): dnnl_status_t; external libdnnl;

/// Returns primitive attributes post-ops.
///
/// @warning
///     The output @p post_ops points to the internal @p attr field, so it is
///     an error to modify or destroy them. The lifetime of @p post_ops is
///     the same as that of the @p attr it belongs to, so it is an error to
///     use @p post_ops after @p attr has been destroyed.
///
function dnnl_primitive_attr_get_post_ops(
        attr: const_dnnl_primitive_attr_t; post_ops: Pconst_dnnl_post_ops_t): dnnl_status_t; external libdnnl;

/// Sets primitive attributes post-ops.
///
function dnnl_primitive_attr_set_post_ops(
        attr: dnnl_primitive_attr_t; post_ops: const_dnnl_post_ops_t): dnnl_status_t; external libdnnl;

/// Creates empty post-ops sequence.
///
function dnnl_post_ops_create(post_ops: Pdnnl_post_ops_t): dnnl_status_t; external libdnnl;

/// Destroys post-ops.
///
function dnnl_post_ops_destroy(post_ops: dnnl_post_ops_t): dnnl_status_t; external libdnnl;

/// Returns the length of post-ops.
///
function dnnl_post_ops_len(post_ops: const_dnnl_post_ops_t): int; external libdnnl;

/// Returns the kind of a post-op entry.
///
function dnnl_post_ops_get_kind(
        post_ops: const_dnnl_post_ops_t; index: int): dnnl_primitive_kind_t; external libdnnl;

/// Appends an accumulation v3 (sum) to post-ops. Prior to accumulating the
/// result, a zero point is subtracted from the previous value and is
/// multiplied by the scale.
///
function dnnl_post_ops_append_sum(post_ops: dnnl_post_ops_t;
        scale: float; zero_point: int32_t; data_type: dnnl_data_type_t): dnnl_status_t; external libdnnl;

/// Appends an elementwise post-op.
///
function dnnl_post_ops_append_eltwise(post_ops: dnnl_post_ops_t;
        scale: float; alg_kind: dnnl_alg_kind_t; alpha, beta: float): dnnl_status_t; external libdnnl;

/// Appends a binary post-op.
///
function dnnl_post_ops_append_binary(post_ops: dnnl_post_ops_t;
        alg_kind: dnnl_alg_kind_t; const src1_desc: Pdnnl_memory_desc_t): dnnl_status_t; external libdnnl;


/// Destroys a memory descriptor.
///
function dnnl_memory_desc_destroy(memory_desc: dnnl_memory_desc_t): dnnl_status_t; external libdnnl;

/// Creates a memory descriptor using dimensions and strides.
///
/// @note
///     As always, the logical order of dimensions corresponds to the `abc...`
///     format tag, and the physical meaning of the dimensions depends on both
///     the primitive that consumes the memory and the context of that
///     consumption.
///
/// @param memory_desc Output memory descriptor.
/// @param ndims Number of dimensions
/// @param dims Array of dimensions.
/// @param data_type Elements data type.
/// @param strides Strides in each dimension.
function dnnl_memory_desc_create_with_strides(
        memory_desc: Pdnnl_memory_desc_t; ndims: int; const dims: dnnl_dims_t;
        data_type: dnnl_data_type_t; const strides: dnnl_dims_t): dnnl_status_t; external libdnnl;

/// Creates a memory descriptor using dimensions and memory format tag.
///
/// @param tag Memory format tag. Can be #dnnl_format_tag_any which would
///     allow a primitive to chose the final memory format. In this case the
///     format_kind field of the memory descriptor would be set to
///     #dnnl_format_kind_any.
function dnnl_memory_desc_create_with_tag(
        memory_desc: Pdnnl_memory_desc_t; ndims: int; const dims: dnnl_dims_t;
        data_type: dnnl_data_type_t; tag: dnnl_format_tag_t): dnnl_status_t; external libdnnl;

/// Creates a memory descriptor for a region inside an area
/// described by an existing memory descriptor.
///
/// @warning
///     Some combinations of physical memory layout and/or offsets or dims may
///     result in a failure to create a submemory.
//
/// @param memory_desc Output memory descriptor.
/// @param parent_memory_desc An existing memory descriptor.
/// @param dims Sizes of the region.
/// @param offsets Offsets to the region from the encompassing
///     memory object in each dimension
function dnnl_memory_desc_create_submemory(
        memory_desc: Pdnnl_memory_desc_t;
        const parent_memory_desc: Pdnnl_memory_desc_t; const dims: dnnl_dims_t;
        const offsets: dnnl_dims_t): dnnl_status_t; external libdnnl;

/// Queries a memory descriptor for various pieces of information.
///
/// The following information can be queried:
///  - Number of dimensions (#dnnl_query_ndims_s32)
///  - Dimensions (#dnnl_query_dims) in the following order:
///    - CNN data tensors: mini-batch, channel, spatial
///      (<code>{N, C, [[D,] H,] W}</code>)
///    - CNN weight tensors: group (optional), output channel, input channel,
///      spatial (<code>{[G,] O, I, [[D,] H,] W}</code>)
///    - RNN data tensors: time, mini-batch, channels (<code>{T, N, C}</code>)
///      or layers, directions, states, mini-batch, channels
///      (<code>{L, D, S, N, C}</code>)
///    - RNN weight tensor: layers, directions, input channel, gates, output
///      channels (<code>{L, D, I, G, O}</code>)
///  - Data type of the tensor elements (#dnnl_query_data_type)
///  - Padded dimensions (#dnnl_query_padded_dims) - size of the data including
///    padding in each dimension
///  - Padded offsets (#dnnl_query_padded_offsets) - per-dimension offset from
///    the padding to actual data, the top-level tensor with offsets applied
///    must lie within the padding area.
///  - Submemory offset (#dnnl_query_submemory_offset_s64) - offset from memory
///    origin to the current block, non-zero only in a description of a memory
///    sub-block.
///  - Format kind (#dnnl_query_format_kind) - memory format kind
///
function dnnl_memory_desc_query(
        memory_desc: const_dnnl_memory_desc_t; what: dnnl_query_t; result: Pointer): dnnl_status_t; external libdnnl;

/// Compares two memory descriptors.
///
/// Use this function to identify whether a reorder is required between the
/// two memories
///
/// @returns 1 if the descriptors are the same, 0 if the descriptors are different.
function dnnl_memory_desc_equal(const lhs, rhs: const_dnnl_memory_desc_t): int; external libdnnl;

/// Returns the size of a memory descriptor.
///
///   The size of data in general cannot be computed as the product of dimensions multiplied by the size of the data type.
///   So users are required to use this function for better code portability.
function dnnl_memory_desc_get_size(const memory_desc: const_dnnl_memory_desc_t): size_t; external libdnnl;

/// Returns the size of data type.
///
/// @param data_type Data type.
/// @returns The number of bytes occupied by data type.
function dnnl_data_type_size(data_type: dnnl_data_type_t): size_t; external libdnnl;

/// Creates a memory object.
///
function dnnl_memory_create(memory: Pdnnl_memory_t; const memory_desc: const_dnnl_memory_desc_t;
        engine: dnnl_engine_t; handle: Pointer): dnnl_status_t; external libdnnl;

/// Returns the memory descriptor for a memory object.
///
function dnnl_memory_get_memory_desc(memory: const_dnnl_memory_t;
        const memory_desc: Pconst_dnnl_memory_desc_t): dnnl_status_t; external libdnnl;

/// Returns the engine of a memory object.
///
function dnnl_memory_get_engine(memory: const_dnnl_memory_t; engine: Pdnnl_engine_t): dnnl_status_t; external libdnnl;

/// Maps a memory object and returns a host-side pointer to a memory buffer
/// with a copy of its contents.
///
/// Mapping enables explicit direct access to memory contents for the engines
/// that do not support it implicitly.
///
/// Mapping is an exclusive operation - a memory object cannot be used in
/// other operations until this memory object is unmapped.
///
/// @note
///     The dnnl_memory_map_data() and dnnl_memory_unmap_data() functions are
///     mainly provided for debug and testing purposes, and their performance
///     may be suboptimal.
///
function dnnl_memory_map_data(
        memory: const_dnnl_memory_t; mapped_ptr: PPointer): dnnl_status_t; external libdnnl;

/// Unmaps a memory object and writes back any changes made to the previously
/// mapped memory buffer. The pointer to the mapped buffer must be obtained
/// via the dnnl_memory_map_data() call.
///
function dnnl_memory_unmap_data(
        memory: const_dnnl_memory_t; mapped_ptr: Pointer): dnnl_status_t; external libdnnl;

/// Returns memory object's data handle.
///
function dnnl_memory_get_data_handle(memory: const_dnnl_memory_t; handle: PPointer): dnnl_status_t; external libdnnl;

/// Sets the underlying memory buffer.
///
function dnnl_memory_set_data_handle(memory: dnnl_memory_t; handle: Pointer): dnnl_status_t; external libdnnl;

/// Destroys a memory object.
///
function dnnl_memory_destroy(memory: dnnl_memory_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a reorder primitive.
///
function dnnl_reorder_primitive_desc_create(
        reorder_primitive_desc: Pdnnl_primitive_desc_t;
        const src_desc: const_dnnl_memory_desc_t; src_engine: dnnl_engine_t;
        const dst_desc: const_dnnl_memory_desc_t; dst_engine: dnnl_engine_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an out-of-place concatenation
/// primitive.
///
function dnnl_concat_primitive_desc_create(
        concat_primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        dst_desc: const_dnnl_memory_desc_t; n: int; concat_dimension: int;
        const src_descs: Pconst_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an (out-of-place) sum primitive.
///
function dnnl_sum_primitive_desc_create(
        sum_primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        dst_desc: const_dnnl_memory_desc_t; n: int; const scales: Pfloat;
        const src_descs: Pconst_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a binary primitive.
///
function dnnl_binary_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        src0_desc, src1_desc, dst_desc: const_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a convolution forward propagation
///     primitive.
/// @note
///     Memory descriptors can be initialized with
///     #dnnl_format_tag_any or with format_kind set to #dnnl_format_kind_any.
///
/// Arrays @p strides, @p dilates, @p padding_l, and @p padding_r contain
/// values for spatial dimensions only and hence must have the same number of
/// elements as there are spatial dimensions. The order of values is the same
/// as in the tensor: depth (for 3D tensors), height (for 3D and 2D tensors),
/// and width.
///
function dnnl_convolution_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, weights_desc, bias_desc, dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;


/// Creates a primitive descriptor for a convolution backward propagation
///     primitive.
///
function dnnl_convolution_backward_data_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, weights_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a convolution weights gradient primitive.
///
function dnnl_convolution_backward_weights_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        src_desc, diff_weights_desc, diff_bias_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a deconvolution forward propagation
///     primitive.
///
function dnnl_deconvolution_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, weights_desc, bias_desc, dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a deconvolution backward propagation
///     primitive.
///
function dnnl_deconvolution_backward_data_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, weights_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a deconvolution weights gradient
///     primitive.
///
function dnnl_deconvolution_backward_weights_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        src_desc, diff_weights_desc, diff_bias_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        const strides, dilates, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a shuffle forward propagation primitive
///
function dnnl_shuffle_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        axis: int; group_size: dnnl_dim_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a shuffle backward propagation primitive
///
function dnnl_shuffle_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        diff_src_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        axis: int; group_size: dnnl_dim_t;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an eltwise forward propagation primitive.
///
/// @param alpha The alpha parameter for the elementwise operation. Specific
///     meaning depends on the algorithm.
/// @param beta The beta parameter for the elementwise operation. Specific
///     meaning depends on the algorithm.
function dnnl_eltwise_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        alpha, beta: float;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an eltwise backward propagation
///     primitive.
function dnnl_eltwise_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, diff_dst_desc, data_desc: const_dnnl_memory_desc_t;
        alpha, beta: float;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a softmax forward propagation primitive.
///
function dnnl_softmax_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        softmax_axis: int;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a softmax backward propagation primitive.
///
function dnnl_softmax_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, diff_dst_desc, dst_desc: const_dnnl_memory_desc_t;
        softmax_axis: int;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a pooling forward propagation
///     primitive.
function dnnl_pooling_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        const strides, kernel, dilation, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a pooling backward propagation
///     primitive.
function dnnl_pooling_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        const strides, kernel, dilation, padding_l, padding_r: Pdnnl_dim_t{dnnl_dims_t};
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a PReLU (leaky ReLU with trainable
///     alpha parameter) forward propagation primitive.
///
function dnnl_prelu_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        src_desc, weights_desc, dst_desc: const_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a PReLU (leaky ReLU with trainable
///     alpha parameter) backward propagation primitive.
///
function dnnl_prelu_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        src_desc, weights_desc, diff_src_desc, diff_weights_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an LRN forward propagation primitive.
///
function dnnl_lrn_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        local_size: dnnl_dim_t; alpha, beta, k: float;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an LRN backward propagation primitive.
///
function dnnl_lrn_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        diff_src_desc, diff_dst_desc, src_desc: const_dnnl_memory_desc_t;
        local_size: dnnl_dim_t; alpha, beta, k: float;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a batch normalization forward propagation
///     primitive.
///
function dnnl_batch_normalization_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        epsilon: float; flags: unsigned;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a batch normalization backward
///     propagation primitive.
///
function dnnl_batch_normalization_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        diff_src_desc, diff_dst_desc, src_desc: const_dnnl_memory_desc_t;
        epsilon: float; flags: unsigned;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a layer normalization forward propagation
///     primitive.
///
function dnnl_layer_normalization_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        src_desc, dst_desc, stat_desc: const_dnnl_memory_desc_t;
        epsilon: float; flags: unsigned;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a layer normalization backward
///     propagation primitive.
///
function dnnl_layer_normalization_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        diff_src_desc, diff_dst_desc, src_desc, stat_desc: const_dnnl_memory_desc_t;
        epsilon: float; flags: unsigned;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an inner product forward propagation
///     primitive.
///
function dnnl_inner_product_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t;
        src_desc, weights_desc, bias_desc, dst_desc: const_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an inner product backward propagation
///     primitive.
///
function dnnl_inner_product_backward_data_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        diff_src_desc, weights_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for an inner product  weights gradient
///     primitive.
///
function dnnl_inner_product_backward_weights_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        src_desc, diff_weights_desc, diff_bias_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a matrix multiplication primitive.
///
function dnnl_matmul_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        src_desc, weights_desc, bias_desc, dst_desc: const_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a resampling forward propagation
///     primitive.
///
function dnnl_resampling_forward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        prop_kind: dnnl_prop_kind_t; alg_kind: dnnl_alg_kind_t;
        const factors: Pfloat;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a resampling backward propagation
///     primitive.
///
function dnnl_resampling_backward_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        const factors: Pfloat;
        diff_src_desc, diff_dst_desc: const_dnnl_memory_desc_t;
        hint_fwd_pd: const_dnnl_primitive_desc_t;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;

/// Creates a primitive descriptor for a reduction primitive.
///
function dnnl_reduction_primitive_desc_create(
        primitive_desc: Pdnnl_primitive_desc_t; engine: dnnl_engine_t;
        alg_kind: dnnl_alg_kind_t;
        src_desc, dst_desc: const_dnnl_memory_desc_t;
        p, eps: float;
        attr: const_dnnl_primitive_attr_t): dnnl_status_t; external libdnnl;


implementation

end.

